@@ -65,7 +65,7 @@ class XGBoost(ModelBase):
         ################################# LVL 2 ########################################
         if self._opt_lvl == 2:
             self.model_param['max_depth'] = trial.suggest_int('xgb_max_depth', 2, 12,)
-            self.model_param['min_child_weight'] = trial.suggest_int('xgb_min_child_weight', 1, (len(self.X_train)//100))
+            self.model_param['min_child_weight'] = trial.suggest_int('xgb_min_child_weight', 1, (len(self._data.X_train)//100))
                        
         if self._opt_lvl >= 2:
             self.model_param['learning_rate'] = trial.suggest_int('xgb_learning_rate', 1, 100)/1000
@@ -77,7 +77,7 @@ class XGBoost(ModelBase):
             self.model_param['booster'] = trial.suggest_categorical('xgb_booster', ['gbtree', 'dart', 'gblinear'])
             
             if self.model_param['booster'] == 'dart' or self.model_param['booster'] == 'gbtree':
-                self.model_param['min_child_weight'] = trial.suggest_int('xg_min_child_weight', 1, (len(self.X_train)//100))
+                self.model_param['min_child_weight'] = trial.suggest_int('xg_min_child_weight', 1, (len(self._data.X_train)//100))
                 self.model_param['max_depth'] = trial.suggest_int('xgb_max_depth', 1, 20)
                 self.model_param['gamma'] = trial.suggest_loguniform('xgb_gamma', 1e-6, 1.0)
                 self.model_param['grow_policy'] = trial.suggest_categorical('xgb_grow_policy', ['depthwise', 'lossguide'])
@@ -111,7 +111,7 @@ class XGBoost(ModelBase):
                 self.model_param['scale_pos_weight'] = trial.suggest_discrete_uniform('xgb_scale_pos_weight', 0.1, 1., 0.1)
 
 
-    def _fit(self, X_train=None, y_train=None, X_test=None, y_test=None,):
+    def _fit(self, model=None, X_train=None, y_train=None, X_test=None, y_test=None,):
         """
         Args:
             X (pd.DataFrame, shape (n_samples, n_features)): the input data
@@ -119,16 +119,19 @@ class XGBoost(ModelBase):
         Return:
             self
         """
-        if (X_train is None) or (y_train is None):
-            X_train = self.X_train
-            y_train = self.y_train
+        if model is None:
+            model = self
 
-        params = self.model_param.copy()
+        if (X_train is None) or (y_train is None):
+            X_train = model._data.X_train
+            y_train = model._data.y_train
+
+        params = model.model_param.copy()
         early_stopping_rounds = params.pop('early_stopping_rounds')
        
-        self.model = self._init_model(model_param=params)
-        if self.wrapper_params['early_stopping'] and (X_test is not None):
-            self.model.fit(
+        model.model = model._init_model(model_param=params)
+        if model.wrapper_params['early_stopping'] and (X_test is not None):
+            model.model.fit(
                 X_train, 
                 y_train,
                 eval_set=(X_test, y_test,),
@@ -136,13 +139,12 @@ class XGBoost(ModelBase):
                 verbose=False,
                 )
         else:
-            self.model.fit(
+            model.model.fit(
                 X_train, 
                 y_train,
                 verbose=False,
                 )
-
-        return self
+        return model
 
     def _predict(self, X=None):
         """
@@ -198,9 +200,7 @@ class XGBoost(ModelBase):
 
 class XGBoostClassifier(XGBoost):
     type_of_estimator='classifier'
-    __name__ = 'XGBoostClassifier'
 
 
 class XGBoostRegressor(XGBoost):
-    type_of_estimator='regression'
-    __name__ = 'XGBoostRegressor'
\ No newline at end of file
+    type_of_estimator='regression'
\ No newline at end of file
